import os
from pathlib import Path
import time
import json
from typing import Callable, Optional, Union

import requests
import urllib3
from urllib.parse import urljoin, urlparse
import logging

from instagrapi.mixins.story import Story
from instagrapi.types import Media
from instagrapi import Client
from instagrapi.exceptions import LoginRequired, PleaseWaitFewMinutes

from warp_beacon.scrapler.abstract import ScraplerAbstract

INST_SESSION_FILE = "/var/warp_beacon/inst_session.json"

class InstagramScrapler(ScraplerAbstract):
	cl = None

	def __init__(self) -> None:
		super().__init__()
		self.cl = Client()

	def safe_write_session(self) -> None:
		tmp_fname = "%s~" % INST_SESSION_FILE
		with open(tmp_fname, 'w+') as f:
			f.write(json.dumps(self.cl.get_settings()))
		if os.path.isfile(INST_SESSION_FILE):
			os.unlink(INST_SESSION_FILE)
		os.rename(tmp_fname, INST_SESSION_FILE)

	def load_session(self) -> None:
		if os.path.exists(INST_SESSION_FILE):
			self.cl.load_settings(INST_SESSION_FILE)
		else:
			self.login()

	def login(self) -> None:
		self.cl = Client()
		username = os.environ.get("INSTAGRAM_LOGIN", default=None)
		password = os.environ.get("INSTAGRAM_PASSWORD", default=None)
		verification_code = os.environ.get("INSTAGRAM_VERIFICATION_CODE", default="")
		if username is not None and password is not None:
			self.cl.login(username=username, password=password, verification_code=verification_code)
		self.safe_write_session()

	def scrap(self, url: str) -> tuple[str]:
		self.load_session()
		def _scrap() -> tuple[str]:
			if "stories" in url:
				# remove URL options
				_url = urljoin(url, urlparse(url).path)
				url_last_part = list(filter(None, _url.split('/')))[-1]
				logging.debug("url last part: '%s'", url_last_part)
				if url_last_part.isnumeric():
					return "story", self.scrap_story(url)
				else:
					return "stories", url_last_part
			else:
				return "media", self.scrap_media(url)
		try:
			return _scrap()
		except LoginRequired as e:
			logging.warning("Session error. Trying to relogin...")
			logging.exception(e)
			self.login()
			return _scrap()

	def scrap_stories(self, username: str) -> list[Story]:
		user_info = self.cl.user_info_by_username(username)
		logging.info("user_id is '%s'", user_info.pk)
		return self.cl.user_stories(user_id=user_info.pk)

	def scrap_story(self, url: str) -> str:
		story_id = self.cl.story_pk_from_url(url)
		logging.info("story_id is '%s'", story_id)
		return story_id

	def scrap_media(self, url: str) -> str:
		media_id = self.cl.media_pk_from_url(url)
		logging.info("media_id is '%s'", media_id)
		return media_id
	
	def __download_hndlr(self, func: Callable, *args: tuple[str], **kwargs: dict[str]) -> Union[Path, Media]:
		ret_val = {}
		max_retries = int(os.environ.get("IG_MAX_RETRIES", default=5))
		retries = 0
		while max_retries >= retries:
			try:
				ret_val = func(*args, **kwargs)
				break
			except (requests.exceptions.ConnectionError,
					requests.exceptions.ReadTimeout,
					urllib3.exceptions.ReadTimeoutError,
					urllib3.exceptions.ConnectionError) as e:
				logging.warning("Instagram read timeout! Retrying in 2 seconds ...")
				logging.info("Your `IG_MAX_RETRIES` values is %d", max_retries)
				logging.exception(e)
				if max_retries == retries:
					raise e
				retries += 1
				time.sleep(2)

		return ret_val

	
	def download_video(self, url: str, media_info: dict) -> dict:
		path = self.__download_hndlr(self.cl.video_download_by_url, url, folder='/tmp')
		return {"local_media_path": str(path), "media_type": "video", "media_info": {"duration": media_info.video_duration}}

	def download_photo(self, url: str) -> dict:
		path = self.__download_hndlr(self.cl.photo_download_by_url, url, folder='/tmp')
		return {"local_media_path": str(path), "media_type": "image"}
	
	def download_story(self, story_info: Story) -> dict:
		path, media_type, media_info = "", "", {}
		logging.info("Story id is '%s'", story_info.id)
		effective_story_id = story_info.id
		if '_' in effective_story_id:
			st_parts = effective_story_id.split('_')
			if len(st_parts) > 1:
				effective_story_id = st_parts[0]
		logging.info("Effective story id is '%s'", effective_story_id)
		effective_url = "https://www.instagram.com/stories/%s/%s/" % (story_info.user.username, effective_story_id)
		if story_info.media_type == 1: # photo
			path = self.__download_hndlr(self.cl.story_download_by_url, url=story_info.thumbnail_url, folder='/tmp')
			media_type = "image"
		elif story_info.media_type == 2: # video
			path = self.__download_hndlr(self.cl.story_download_by_url, url=story_info.video_url, folder='/tmp')
			media_type = "video"
			media_info["duration"] = story_info.video_duration

		return {"local_media_path": str(path), "media_type": media_type, "media_info": media_info, "effective_url": effective_url}

	def download_stories(self, stories: list[Story]) -> dict:
		res = []
		for story in stories:
			res.append(self.download_story(story_info=story))

		return {"media_type": "collection", "save_items": True, "items": res}

	def download_album(self, media_info: dict) -> dict:
		res = []
		for i in media_info.resources:
			_media_info = self.cl.media_info(i.pk)
			if i.media_type == 1: # photo
				res.append(self.download_photo(url=_media_info.thumbnail_url))
			elif i.media_type == 2: # video
				res.append(self.download_video(url=_media_info.video_url, media_info=_media_info))

		return {"media_type": "collection", "items": res}

	def download(self, url: str) -> Optional[list[dict]]:
		res = []
		while True:
			try:
				scrap_type, media_id = self.scrap(url)
				if scrap_type == "media":
					media_info = self.__download_hndlr(self.cl.media_info, media_id)
					logging.info("media_type is '%d', product_type is '%s'", media_info.media_type, media_info.product_type)
					if media_info.media_type == 2 and media_info.product_type == "clips": # Reels
						res.append(self.download_video(url=media_info.video_url, media_info=media_info))
					elif media_info.media_type == 1: # Photo
						res.append(self.download_photo(url=media_info.thumbnail_url))
					elif media_info.media_type == 8: # Album
						res.append(self.download_album(media_info=media_info))
				elif scrap_type == "story":
					story_info = self.cl.story_info(media_id)
					logging.info("media_type for story is '%d'", story_info.media_type)
					res.append(self.download_story(story_info=story_info))
				elif scrap_type == "stories":
					logging.info("Stories download mode")
					res.append(self.download_stories(self.scrap_stories(media_id)))
				break
			except PleaseWaitFewMinutes as e:
				logging.warning("Please wait a few minutes error. Trying to relogin ...")
				logging.exception(e)
				wait_timeout = int(os.environ.get("IG_WAIT_TIMEOUT", default=5))
				logging.info("Waiting %d seconds according configuration option `IG_WAIT_TIMEOUT`", wait_timeout)
				if res:
					for i in res:
						if i["media_type"] == "collection":
							for j in i["items"]:
								if os.path.exists(j["local_media_path"]):
									os.unlink(j["local_media_path"])
						else:
							if os.path.exists(i["local_media_path"]):
								os.unlink(i["local_media_path"])
				os.unlink(INST_SESSION_FILE)
				time.sleep(wait_timeout)
		return res



import os
import time

from typing import Optional
import multiprocessing
from requests.exceptions import ConnectTimeout, HTTPError
from instagrapi.exceptions import MediaNotFound, UnknownError, ClientNotFoundError, UserNotFound

from warp_beacon.mediainfo.video import VideoInfo
from warp_beacon.compress.video import VideoCompress
from warp_beacon.uploader import AsyncUploader
from warp_beacon.jobs.download_job import DownloadJob

import logging

CONST_CPU_COUNT = multiprocessing.cpu_count()

class AsyncDownloader(object):
	workers = []
	allow_loop = None
	job_queue = multiprocessing.Queue()
	uploader = None
	workers_count = CONST_CPU_COUNT

	def __init__(self, uploader: AsyncUploader, workers_count: int=CONST_CPU_COUNT) -> None:
		self.allow_loop = multiprocessing.Value('i', 1)
		self.uploader = uploader
		self.workers_count = workers_count

	def __del__(self) -> None:
		self.stop_all()

	def start(self) -> None:
		for _ in range(self.workers_count):
			proc = multiprocessing.Process(target=self.do_work)
			self.workers.append(proc)
			proc.start()

	def get_media_info(self, path: str, fr_media_info: dict={}) -> Optional[dict]:
		media_info = None
		try:
			if path:
				video_info = VideoInfo(path)
				media_info = video_info.get_finfo(tuple(fr_media_info.keys()))
				media_info.update(fr_media_info)
				media_info["thumb"] = video_info.generate_thumbnail()
				logging.info("Media file info: %s", media_info)
		except Exception as e:
			logging.error("Failed to process media info!")
			logging.exception(e)

		return media_info

	def do_work(self) -> None:
		logging.info("download worker started")
		while self.allow_loop.value == 1:
			try:
				job = None
				try:
					job = self.job_queue.get()
					actor = None
					try:
						items = []
						if "instagram.com/" in job.url:
							if not job.in_process:
								from warp_beacon.scrapler.instagram import InstagramScrapler
								actor = InstagramScrapler()
								while True:
									try:
										logging.info("Downloading URL '%s'", job.url)
										items = actor.download(job.url)
										break
									except ConnectTimeout as e:
										logging.error("ConnectTimeout download error!")
										logging.exception(e)
										time.sleep(2)
									except (MediaNotFound, ClientNotFoundError, UserNotFound) as e:
										logging.warning("Not found error occurred!")
										logging.exception(e)
										self.uploader.queue_task(job.to_upload_job(
											job_failed=True,
											job_failed_msg="Unable to access to media under this URL. Seems like the media is private.")
										)
										break
									except (UnknownError, Exception) as e:
										logging.warning("UnknownError occurred!")
										logging.exception(e)
										exception_msg = ""
										if hasattr(e, "message"):
											exception_msg = e.message
										else:
											exception_msg = str(e)
										if "geoblock_required" in exception_msg:
											self.uploader.queue_task(job.to_upload_job(
												job_failed=True,
												job_failed_msg="This content does not accessible for bot account. Seems like author blocked certain region.")
											)
											break
										self.uploader.queue_task(job.to_upload_job(
											job_failed=True,
											job_failed_msg="WOW, unknown error occured! Please send service logs to developer via email: andrey@bagrintsev.me.")
										)
										break

								if items:
									for item in items:
										media_info = {"filesize": 0}
										if item["media_type"] == "video":
											media_info = self.get_media_info(item["local_media_path"], item["media_info"])
											if media_info["filesize"] > 50.0:
												logging.info("Detected big file. Starting compressing with ffmpeg ...")
												self.uploader.queue_task(job.to_upload_job(
													job_warning=True,
													job_warning_msg="Downloaded file size is bigger than Telegram limits\! Performing video compression\. This may take a while\.")
												)
												ffmpeg = VideoCompress(file_path=item["local_media_path"])
												new_filepath = ffmpeg.generate_filepath(base_filepath=item["local_media_path"])
												if ffmpeg.compress_to(new_filepath, target_size=50 * 1000):
													logging.info("Successfully compressed file '%s'", new_filepath)
													os.unlink(item["local_media_path"])
													item["local_media_path"] = new_filepath
													item["local_compressed_media_path"] = new_filepath
													media_info["filesize"] = VideoInfo.get_filesize(new_filepath)
													logging.info("New file size of compressed file is '%.3f'", media_info["filesize"])
										elif item["media_type"] == "collection":
											for v in item["items"]:
												if v["media_type"] == "video":
													col_media_info = self.get_media_info(v["local_media_path"], v["media_info"])
													media_info["filesize"] += int(col_media_info.get("filesize", 0))
													v["media_info"] = col_media_info

										job_args = {"media_type": item["media_type"], "media_info": media_info}
										if item["media_type"] == "collection":
											job_args["media_collection"] = item["items"]
											if item.get("save_items", None) is not None:
												job_args["save_items"] = item.get("save_items", False)
										else:
											job_args["local_media_path"] = item["local_media_path"]
											if item.get("local_compressed_media_path", None):
												job_args["local_media_path"] = item.get("local_compressed_media_path", None)

										logging.debug("local_media_path: '%s'", job_args.get("local_media_path", ""))
										logging.debug("media_collection: '%s'", str(job_args.get("media_collection", {})))
										upload_job = job.to_upload_job(**job_args)
										if upload_job.is_empty():
											logging.info("Upload job is empty. Nothing to do here!")
											self.uploader.queue_task(job.to_upload_job(
												job_failed=True,
												job_failed_msg="Seems like this link doesn't contains any media.")
											)
										else:
											self.uploader.queue_task(upload_job)
							else:
								logging.info("Job already in work in parallel worker. Redirecting job to upload worker.")
								self.uploader.queue_task(job.to_upload_job())
					except HTTPError as e:
						logging.error("HTTP error inside download worker!")
						logging.exception(e)
					except Exception as e:
						logging.error("Error inside download worker!")
						logging.exception(e)
						self.notify_task_failed(job)
						#self.queue_task(url=item["url"], message_id=item["message_id"], item_in_process=item["in_process"], uniq_id=item["uniq_id"])
				except multiprocessing.Queue.empty:
					pass
			except Exception as e:
				logging.error("Exception occurred inside worker!")
				logging.exception(e)

	def stop_all(self) -> None:
		self.allow_loop.value = 0
		for proc in self.workers:
			if proc.is_alive():
				logging.info("stopping process #%d", proc.pid)
				proc.terminate()
				#proc.join()
				logging.info("process #%d stopped", proc.pid)
		self.workers.clear()

	def queue_task(self, job: DownloadJob) -> str:
		self.job_queue.put_nowait(job)
		return str(job.job_id)
	
	def notify_task_failed(self, job: DownloadJob) -> None:
		self.uploader.queue_task(job.to_upload_job(job_failed=True))


from typing import TypedDict
from typing_extensions import Unpack

from warp_beacon.jobs.upload_job import UploadJob
from warp_beacon.jobs.abstract import AbstractJob, JobSettings

class DownloadJob(AbstractJob):
	def __init__(self, **kwargs: Unpack[JobSettings]) -> None:
		super(DownloadJob, self).__init__(**kwargs)
	def __del__(self) -> None:
		pass

	@staticmethod
	def build(**kwargs: Unpack[JobSettings]) -> "DownloadJob":
		return DownloadJob(**kwargs)
	
	def to_upload_job(self, **kwargs: Unpack[JobSettings]) -> AbstractJob:
		d = self.to_dict()
		d.update(kwargs)
		if "media_collection" in d:
			for k, v in enumerate(d["media_collection"]):
				d["media_collection"][k] = UploadJob.build(**v)
		return UploadJob.build(**d)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import signal
import asyncio
import time
from io import BytesIO
import logging

from urlextract import URLExtract

import telegram
from telegram import Bot, ForceReply, Update, Chat, error, InputMediaVideo, InputMediaPhoto, MessageEntity, InlineKeyboardMarkup, InlineKeyboardButton
from telegram.ext import Application, CommandHandler, ContextTypes, MessageHandler, filters
from telegram.constants import ParseMode

import warp_beacon.scrapler
from warp_beacon.storage import Storage
from warp_beacon.uploader import AsyncUploader
from warp_beacon.jobs.download_job import DownloadJob, UploadJob

# Enable logging
logging.basicConfig(
	format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
# set higher logging level for httpx to avoid all GET and POST requests being logged
logging.getLogger("httpx").setLevel(logging.WARNING)

logger = logging.getLogger(__name__)

storage = Storage()
uploader = None
downloader = None
placeholder = ("animation", None)

# Define a few command handlers. These usually take the two arguments update and
# context.
async def start(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
	"""Send a message when the command /start is issued."""
	user = update.effective_user
	await update.message.reply_html(
		rf"Hi {user.mention_html()}!",
		reply_markup=ForceReply(selective=True),
	)

async def help_command(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
	"""Send a message when the command /help is issued."""
	await update.message.reply_text("Send me a link to remote media")

async def random(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
	d = storage.get_random()
	if not d:
		await update.message.reply_text("No random content yet.")
		return
	await upload_job(update, context, UploadJob(tg_file_id=d["tg_file_id"], media_type=d["media_type"], message_id=update.message.message_id))

async def remove_placeholder(update: Update, context: ContextTypes.DEFAULT_TYPE, placeholder_message_id: int) -> None:
	try:
		timeout = int(os.environ.get("TG_WRITE_TIMEOUT", default=120))
		deleted_message = await context.bot.delete_message(
			chat_id=update.message.chat_id,
			message_id=placeholder_message_id,
			write_timeout=timeout,
			read_timeout=timeout,
			connect_timeout=timeout
		)
	except Exception as e:
		logging.error("Failed to remove placeholder message!")
		logging.exception(e)

async def update_placeholder_text(update: Update, context: ContextTypes.DEFAULT_TYPE, placeholder_message_id: int, placeholder_text: str) -> None:
	try:
		timeout = int(os.environ.get("TG_WRITE_TIMEOUT", default=120))
		await context.bot.edit_message_caption(
			chat_id=update.message.chat_id,
			message_id=placeholder_message_id,
			parse_mode="MarkdownV2",
			caption=" ⚠️ *%s*" % placeholder_text,
			show_caption_above_media=True,
			write_timeout=timeout,
			read_timeout=timeout,
			connect_timeout=timeout
		)
	except Exception as e:
		logging.error("Failed to update placeholder message!")
		logging.exception(e)

async def send_text(update: Update, context: ContextTypes.DEFAULT_TYPE, reply_id: int, text: str) -> int:
	try:
		reply = await update.message.reply_text(
			text,
			reply_to_message_id=reply_id
		)

		return reply.message_id
	except Exception as e:
		logging.error("Failed to send text message!")
		logging.exception(e)

	return 0

def create_default_placeholder_img(text: str, width: int = 800, height: int = 1280) -> BytesIO:
	from PIL import Image, ImageDraw, ImageFont
	bio = BytesIO()
	bio.name = 'placeholder.png'
	img = Image.new("RGB", (width, height), (255, 255, 255))
	draw = ImageDraw.Draw(img)
	font = ImageFont.load_default(size=48)
	_, _, w, h = draw.textbbox((0, 0), text, font=font)
	draw.text(((width-w)/2, (height-h)/2), text, font=font, fill="#000")
	img.save(bio, 'PNG')
	bio.seek(0)
	return bio

def extract_file_id(message: "Message") -> str:
	if message.animation:
		return message.animation.file_id
	if message.photo:
		return message.photo[-1].file_id
	if message.document:
		return message.document.file_id
	if message.video:
		return message.video.file_id

	return None

async def create_placeholder_message(update: Update, context: ContextTypes.DEFAULT_TYPE, reply_id: int) -> int:
	global placeholder
	retry_amount = 0
	max_retries = int(os.environ.get("TG_MAX_RETRIES", default=5))
	while not retry_amount >= max_retries:
		try:
			text = "*Loading, this may take a moment \.\.\.* ⏱️ "
			reply = None
			if placeholder[1] is None:
				ph_found = False
				for ph in ('/var/warp_beacon/placeholder.gif', "%s/placeholder.gif" % os.path.dirname(os.path.abspath(warp_beacon.__file__))):
					if not os.path.exists(ph):
						continue
					reply = await update.message.reply_animation(
						caption=text,
						parse_mode="MarkdownV2",
						show_caption_above_media=True,
						animation=open(ph, 'rb'),
						reply_to_message_id=reply_id
					)
					placeholder = ("animation", extract_file_id(reply))
					ph_found = True
				if not ph_found:
					try:
						reply = await update.message.reply_animation(
							caption=text,
							parse_mode="MarkdownV2",
							show_caption_above_media=True,
							animation="https://bagrintsev.me/warp_beacon/placeholder_that_we_deserve.mp4",
							reply_to_message_id=reply_id
						)
						placeholder = ("animation", extract_file_id(reply))
					except Exception as e:
						logging.error("Failed to download secret placeholder!")
						logging.exception(e)
						img = create_default_placeholder_img("Loading, this may take a moment ...")
						reply = await update.message.reply_photo(
							caption=text,
							show_caption_above_media=True,
							parse_mode="MarkdownV2",
							filename="placeholder.png",
							photo=img,
							reply_to_message_id=reply_id
						)
						placeholder = ("photo", extract_file_id(reply))
			else:
				if placeholder[0] == "animation":
					reply = await update.message.reply_animation(
						caption=text,
						parse_mode="MarkdownV2",
						show_caption_above_media=True,
						animation=placeholder[1],
						reply_to_message_id=reply_id
					)
				else:
					reply = await update.message.reply_photo(
						caption=text,
						parse_mode="MarkdownV2",
						show_caption_above_media=True,
						photo=placeholder[1],
						reply_to_message_id=reply_id
					)
			return reply.message_id
		except telegram.error.RetryAfter as e:
			logging.warning("RetryAfter exception!")
			logging.exception(e)
			await send_text(update, context, None, "Telegram error: %s" % e.message)
			time.sleep(e.retry_after)
		except Exception as e:
			logging.error("Failed to create placeholder message!")
			logging.exception(e)
			retry_amount += 1
			time.sleep(2)

	return 0

def build_tg_args(update: Update, context: ContextTypes.DEFAULT_TYPE, job: UploadJob) -> dict:
	args = {}
	timeout = int(os.environ.get("TG_WRITE_TIMEOUT", default=120))
	if job.media_type == "video":
		if job.tg_file_id:
			if job.placeholder_message_id:
				args["media"] = InputMediaVideo(media=job.tg_file_id.replace(":video", ''), supports_streaming=True)
			else:
				args["video"] = job.tg_file_id.replace(":video", '')
		else:
			args["media"] = InputMediaVideo(
				media=open(job.local_media_path, 'rb'),
				supports_streaming=True,
				width=job.media_info["width"],
				height=job.media_info["height"],
				duration=int(job.media_info["duration"]),
				thumbnail=job.media_info["thumb"],
				filename="downloaded_via_warp_beacon_bot%s" % (os.path.splitext(job.local_media_path)[-1])
			)
	elif job.media_type == "image":
		if job.tg_file_id:
			if job.placeholder_message_id:
				args["media"] = InputMediaPhoto(media=job.tg_file_id.replace(":image", ''))
			else:
				args["photo"] = job.tg_file_id.replace(":image", '')
		else:
			args["media"] = InputMediaPhoto(
				media=open(job.local_media_path, 'rb'),
				filename="downloaded_via_warp_beacon_bot%s" % (os.path.splitext(job.local_media_path)[-1])
			)
	elif job.media_type == "collection":
		if job.tg_file_id:
			args["media"] = []
			for i in job.tg_file_id.split(','):
				tg_id, mtype = i.split(':')
				ptr = None
				if mtype == "video":
					ptr = InputMediaVideo(media=tg_id)
				elif mtype == "image":
					ptr = InputMediaPhoto(media=tg_id)
				args["media"].append(ptr)
		else:
			mediafs = []
			for j in job.media_collection:
				if j.media_type == "video":
					vid = InputMediaVideo(
						media=open(j.local_media_path, 'rb'),
						supports_streaming=True,
						width=j.media_info["width"],
						height=j.media_info["height"],
						duration=int(j.media_info["duration"]),
						thumbnail=j.media_info["thumb"],
						filename="downloaded_via_warp_beacon_bot%s" % (os.path.splitext(j.local_media_path)[-1])
					)
					mediafs.append(vid)
				elif j.media_type == "image":
					photo = InputMediaPhoto(
						media=open(j.local_media_path, 'rb'),
						filename="downloaded_via_warp_beacon_bot%s" % (os.path.splitext(job.local_media_path)[-1])
					)
					mediafs.append(photo)
			args["media"] = mediafs

	# common args
	if job.placeholder_message_id and job.media_type != "collection":
		args["message_id"] = job.placeholder_message_id
		args["chat_id"] = update.message.chat_id
	else:
		args["disable_notification"] = True
		args["reply_to_message_id"] = job.message_id
	args["write_timeout"] = timeout
	args["read_timeout"] = timeout
	args["connect_timeout"] = timeout
	if os.environ.get("ENABLE_DONATES", None) == "true" and job.media_type != "collection":
		args["reply_markup"] = InlineKeyboardMarkup([[InlineKeyboardButton("❤ Donate", url=os.environ.get("DONATE_LINK", "https://pay.cryptocloud.plus/pos/W5BMtNQt5bJFoW2E"))]])
	return args

async def upload_job(update: Update, context: ContextTypes.DEFAULT_TYPE, job: UploadJob) -> list[str]:
	timeout = int(os.environ.get("TG_WRITE_TIMEOUT", default=120))
	tg_file_ids = []
	try:
		retry_amount = 0
		max_retries = int(os.environ.get("TG_MAX_RETRIES", default=5))
		while not retry_amount >= max_retries:
			try:
				message = None
				if job.media_type == "video":
					if job.placeholder_message_id:
						message = await context.bot.edit_message_media(**build_tg_args(update, context, job))
					else:
						message = await update.message.reply_video(**build_tg_args(update, context, job))
					tg_file_ids.append(message.video.file_id)
					job.tg_file_id = message.video.file_id
					logging.info("Uploaded video file tg_file_id is '%s'", job.tg_file_id)
				elif job.media_type == "image":
					if job.placeholder_message_id:
						message = await context.bot.edit_message_media(**build_tg_args(update, context, job))
					else:
						message = await update.message.reply_photo(**build_tg_args(update, context, job))
					if message.photo:
						tg_file_ids.append(message.photo[-1].file_id)
						job.tg_file_id = message.photo[-1].file_id
				elif job.media_type == "collection":
					sent_messages = await update.message.reply_media_group(**build_tg_args(update, context, job))
					if job.placeholder_message_id:
						await remove_placeholder(update, context, job.placeholder_message_id)
					for i, msg in enumerate(sent_messages):
						if msg.video:
							tg_file_ids.append(msg.video.file_id + ':video')
							if job.media_collection:
								job.media_collection[i].tg_file_id = msg.video.file_id + ':video'
						elif msg.photo:
							tg_file_ids.append(msg.photo[-1].file_id + ':image')
							if job.media_collection:
								job.media_collection[i].tg_file_id = msg.photo[-1].file_id + ':image'
				logging.info("Uploaded to Telegram")
				break
			except (error.NetworkError, error.TimedOut) as e:
				logging.error("Failed to upload due telegram limitations :(")
				logging.exception(e)
				if not "Request Entity Too Large" in e.message:
					logging.info("TG upload will be retried. Configuration `TG_MAX_RETRIES` values is %d.", max_retries)
				
				if "Message to reply not found" in e.message:
					logging.warning("No message to reply found. Looks like original message was deleted by author.")
					job.message_id = None
					continue

				if retry_amount+1 >= max_retries or "Request Entity Too Large" in e.message:
					msg = ""
					if e.message:
						msg = "Telegram error: %s" % str(e.message)
					else:
						msg = "Unfortunately, Telegram limits were exceeded. Your video size is %.2f MB." % job.media_info["filesize"]
					await remove_placeholder(update, context, job.placeholder_message_id)
					await send_text(
						update,
						context,
						job.message_id,
						msg
					)
					break
			retry_amount += 1
	except Exception as e:
		logging.error("Error occurred!")
		logging.exception(e)
	finally:
		if job.media_type == "collection":
			for j in job.media_collection:
				if os.path.exists(j.local_media_path):
					os.unlink(j.local_media_path)
		else:
			if os.path.exists(job.local_media_path):
				os.unlink(job.local_media_path)
		if job.local_compressed_media_path:
			if os.path.exists(job.local_compressed_media_path):
				os.unlink(job.local_compressed_media_path)

	return tg_file_ids

async def handler(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
	if update.message is None:
		return
	chat = update.effective_chat
	effective_message_id = update.message.message_id
	extractor = URLExtract()
	urls = extractor.find_urls(update.message.text_html)

	reply_text = "Wut?"
	if not urls:
		reply_text = "Your message should contains URLs"
	else:
		for url in urls:
			if "instagram.com" not in url:
				logging.info("Only instagram.com is now supported. Skipping.")
				continue
			entities, tg_file_ids = [], []
			uniq_id = Storage.compute_uniq(url)
			try:
				entities = storage.db_lookup_id(uniq_id)
			except Exception as e:
				logging.error("Failed to search link in DB!")
				logging.exception(e)
			if entities:
				tg_file_ids = [i["tg_file_id"] for i in entities]
				logging.info("URL '%s' is found in DB. Sending with tg_file_ids = '%s'", url, str(tg_file_ids))
				ent_len = len(entities)
				if ent_len > 1:
					await upload_job(
						update,
						context,
						UploadJob(
							tg_file_id=",".join(tg_file_ids),
							message_id=effective_message_id,
							media_type="collection"
						)
					)
				elif ent_len:
					media_type = entities.pop()["media_type"]
					await upload_job(
						update,
						context,
						UploadJob(
							tg_file_id=tg_file_ids.pop(),
							message_id=effective_message_id,
							media_type=media_type
						)
					)
			else:
				async def upload_wrapper(job: UploadJob) -> None:
					try:
						if job.job_failed and job.job_failed_msg:
							if job.placeholder_message_id:
								await remove_placeholder(update, context, job.placeholder_message_id)
							return await send_text(update, context, reply_id=job.message_id, text=job.job_failed_msg)
						if job.job_warning and job.job_warning_msg:
							return await update_placeholder_text(update, context, job.placeholder_message_id, job.job_warning_msg)
						tg_file_ids = await upload_job(update, context, job)
						if tg_file_ids:
							if job.media_type == "collection" and job.save_items:
								for i in job.media_collection:
									storage.add_media(tg_file_ids=[i.tg_file_id], media_url=i.effective_url, media_type=i.media_type, origin="instagram")
							else:
								storage.add_media(tg_file_ids=[','.join(tg_file_ids)], media_url=job.url, media_type=job.media_type, origin="instagram")
					except Exception as e:
						logging.error("Exception occurred while performing upload callback!")
						logging.exception(e)

				try:
					# create placeholder message for long download
					placeholder_message_id = await create_placeholder_message(
						update,
						context,
						reply_id=effective_message_id
					)

					if not placeholder_message_id:
						await send_text(
							update=update, 
							context=context, 
							reply_id=effective_message_id, 
							text="Failed to create message placeholder. Please check your bot Internet connection.")
						return

					uploader.add_callback(
						placeholder_message_id,
						upload_wrapper,
						update,
						context
					)

					downloader.queue_task(DownloadJob.build(
						url=url,
						placeholder_message_id=placeholder_message_id,
						message_id=effective_message_id,
						in_process=uploader.is_inprocess(uniq_id),
						uniq_id=uniq_id
					))
					uploader.set_inprocess(uniq_id)
				except Exception as e:
					logging.error("Failed to schedule download task!")
					logging.exception(e)

	if chat.type not in (Chat.GROUP, Chat.SUPERGROUP) and not urls:
		await update.message.reply_text(reply_text, reply_to_message_id=effective_message_id)

@staticmethod
def _raise_system_exit() -> None:
	raise SystemExit

def main() -> None:
	"""Start the bot."""
	try:
		global uploader, downloader

		loop = asyncio.get_event_loop()

		uploader = AsyncUploader(
			storage=storage,
			pool_size=int(os.environ.get("UPLOAD_POOL_SIZE", default=warp_beacon.scrapler.CONST_CPU_COUNT)),
			loop=loop
		)
		downloader = warp_beacon.scrapler.AsyncDownloader(
			workers_count=int(os.environ.get("WORKERS_POOL_SIZE", default=warp_beacon.scrapler.CONST_CPU_COUNT)),
			uploader=uploader
		)
		downloader.start()
		uploader.start()

		stop_signals = (signal.SIGINT, signal.SIGTERM, signal.SIGABRT)
		for sig in stop_signals or []:
			loop.add_signal_handler(sig, _raise_system_exit)
		loop.add_signal_handler(sig, _raise_system_exit)

		# Create the Application and pass it your bot's token.
		tg_token = os.environ.get("TG_TOKEN", default=None)
		application = Application.builder().token(tg_token).concurrent_updates(True).build()

		# on different commands - answer in Telegram
		application.add_handler(CommandHandler("start", start))
		application.add_handler(CommandHandler("random", random))
		application.add_handler(CommandHandler("help", help_command))

		# on non command i.e message - echo the message on Telegram
		application.add_handler(MessageHandler(filters.TEXT & ~filters.COMMAND, handler))

		allow_loop = True
		while allow_loop:
			try:
				loop.run_until_complete(application.initialize())
				if application.post_init:
					loop.run_until_complete(application.post_init(application))
				loop.run_until_complete(application.updater.start_polling())
				loop.run_until_complete(application.start())
				while allow_loop:
					try:
						loop.run_forever()
					except (KeyboardInterrupt, SystemExit) as e:
						allow_loop = False
						raise e
					except Exception as e:
						logging.error("Main loop Telegram error!")
						logging.exception(e)
			except (KeyboardInterrupt, SystemExit):
				logging.debug("Application received stop signal. Shutting down.")
			except telegram.error.TimedOut as e:
				logging.error("Telegram connection timeout!")
				logging.exception(e)
				time.sleep(2)
				continue
			except Exception as e:
				logging.error("Failed to start application!")
				logging.exception(e)
			finally:
				try:
					if application.updater.running:  # type: ignore[union-attr]
						loop.run_until_complete(application.updater.stop())  # type: ignore[union-attr]
					if application.running:
						loop.run_until_complete(application.stop())
					if application.post_stop:
						loop.run_until_complete(application.post_stop(application))
					loop.run_until_complete(application.shutdown())
					if application.post_shutdown:
						loop.run_until_complete(application.post_shutdown(application))
				finally:
					downloader.stop_all()
					uploader.stop_all()
					loop.close()
	except Exception as e:
		logging.exception(e)

if __name__ == "__main__":
	main()



import io, os

from typing import Union
from PIL import Image
import av

import logging

class VideoInfo(object):
	width = 0
	height = 0
	duration = 0.0
	ffmpeg = None
	filename = ""

	def __init__(self, filename: str) -> None:
		self.filename = filename
		with av.open(file=self.filename, mode='r') as container:
			stream = container.streams.video[0]
			time_base = stream.time_base
			self.duration = float(stream.duration * time_base)
			framerate = stream.average_rate
			frame_container_pts = round((1 / framerate) / time_base)
			container.seek(frame_container_pts, backward=True, stream=stream)
			frame = next(container.decode(video=0))
			self.width = frame.width
			self.height = frame.height
		
	def __del__(self) -> None:
		pass

	def get_demensions(self) -> dict:
		return {"width": self.width, "height": self.height}

	def get_duration(self) -> float:
		return self.duration

	@staticmethod
	def get_filesize(filename: str) -> float:
		return os.stat(filename).st_size / 1024 / 1024
	
	def get_finfo(self, except_info: tuple=()) -> dict:
		res = {}
		res.update(self.get_demensions())
		if "duration" not in except_info:
			res["duration"] = int(self.get_duration())
		if "filesize" not in except_info:
			res["filesize"] = round(VideoInfo.get_filesize(self.filename), 2)
		return res
	
	def shrink_image_to_fit(self, image: Image, size: tuple = (320, 320)) -> Image:
		image.thumbnail(size, Image.Resampling.LANCZOS)
		#image.save(
		#	"/tmp/test.th.jpg",
		#	quality=80,
		#)
		return image
	
	def generate_thumbnail(self) -> Union[io.BytesIO, None]:
		try:
			image = None
			with av.open(file=self.filename, mode='r') as container:
				# Signal that we only want to look at keyframes.
				stream = container.streams.video[0]
				stream.codec_context.skip_frame = "NONKEY"
				frame_num = 10
				time_base = container.streams.video[0].time_base
				framerate = container.streams.video[0].average_rate
				frame_container_pts = round((frame_num / framerate) / time_base)
					
				container.seek(frame_container_pts, backward=True, stream=container.streams.video[0])
				frame = next(container.decode(stream))
				
				image = frame.to_image()
				#image.save(
				#	"/tmp/test.{:04d}.jpg".format(frame.pts),
				#	quality=80,
	 			#)
					#break
			if image:
				image = self.shrink_image_to_fit(image)
				io_buf = io.BytesIO()
				image.save(io_buf, format='JPEG')
				io_buf.seek(0)
				return io_buf
		except Exception as e:
			logging.error("Failed to generate thumbnail!")
			logging.exception(e)

		return None



from typing import TypedDict
from typing_extensions import Unpack

from warp_beacon.jobs.abstract import AbstractJob, JobSettings

class UploadJob(AbstractJob):
	def __init__(self, **kwargs: Unpack[JobSettings]) -> None:
		super(UploadJob, self).__init__(**kwargs)
	def __del__(self) -> None:
		pass

	@staticmethod
	def build(**kwargs: Unpack[JobSettings]) -> "UploadJob":
		return UploadJob(**kwargs)
	
	def to_download_job(self, **kwargs: Unpack[JobSettings]) -> AbstractJob:
		from warp_beacon.jobs.download_job import DownloadJob
		d = self.to_dict()
		d.update(kwargs)
		return DownloadJob.build(**d)
	
	def set_flag(self, key: str, value: bool) -> "UploadJob":
		if key in self.__dict__:
			self.__dict__[key] = value

		return self



from abc import ABC, abstractmethod

class ScraplerAbstract(ABC):
	def __init__(self) -> None:
		pass

	def __del__(self) -> None:
		pass

	@abstractmethod
	def scrap(self, url: str) -> str:
		raise NotImplementedError

	@abstractmethod
	def download(self, url: str) -> bool:
		raise NotImplementedError
